/**
  *  \file client/actions/gascript.cc
  *  brief Global Actions: Script Command
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions/gascript.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/ship.h"
#include "int/bytecode.h"
#include "int/csource.h"
#include "int/exec.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/intglobal.h"
#include "int/process.h"
#include "int/statement.h"
#include "ui/msgbox.h"
#include "ui/widgets/inputline.h"
#include "util/nls.h"

class WGlobalScriptAction::Executer : public WGlobalActionExecuter {
    IntBCORef bco;
 public:
    Executer(IntBCORef bco)
        : bco(bco)
        { }
    bool execute(GObject& obj, bool /*lock*/);
    void finish(int obj, int success);
};

WGlobalActionExecuter*
WGlobalScriptAction::prepare()
{
    string_t command;
    while (1) {
        /* Ask for command */
        if (!UIInputLine::run(command, command.max_size(), _("Global Actions"), _("Command:"), 40, 0)) {
            return 0;
        }

        /* Try to compile it */
        try {
            // Make bytecode object
            IntBCORef bco = new IntBytecodeObject();

            // Make command source containing this text
            IntMemoryCommandSource mcs(command);

            // Make compilation context; configure for one-liners
            IntStatementCompilationContext scc;
            scc.withFlags(IntCompilationContext::RefuseBlocks);
            scc.withFlags(IntCompilationContext::LinearExecution);
            scc.withFlags(IntCompilationContext::ExpressionsAreStatements);

            // Compile
            IntStatementCompiler(mcs).compile(*bco, scc);
            if (bco->getNumInstructions() == 0) {
                // blank input
                return 0;
            }

            // We have successfully compiled the input.
            bco->relocate();
            return new Executer(bco);
        }
        catch (std::exception& e) {
            // This includes IntError which is derived from std::exception
            messageBox(format(_("Error compiling this statement: %s"), e.what()), _("Global Actions"));
        }
    }
}

bool
WGlobalScriptAction::Executer::execute(GObject& obj, bool /*lock*/)
{
    // Get object's context
    std::auto_ptr<IntContext> ctx;
    if (GShip* sh = dynamic_cast<GShip*>(&obj)) {
        ctx.reset(new IntShipContext(sh->getId()));
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(&obj)) {
        ctx.reset(new IntPlanetContext(pl->getId()));
    }

    // Operate
    bool result = false;
    if (ctx.get()) {
        try {
            // Create a process
            IntExecutionContext& exc = createProcess("Global Actions");
            exc.pushContextsFrom(global_contexts);
            exc.pushNewContext(ctx.release());
            exc.markContextTOS();
            exc.pushNewFrame(new IntExecutionFrame(bco));
            runRunnableProcesses();
            switch (exc.getState()) {
             case IntExecutionContext::psSuspended:    // 'Stop'
             case IntExecutionContext::psEnded:        // 'End'
             case IntExecutionContext::psTerminated:   // successful execution
             case IntExecutionContext::psTerminating:  // can't happen, but otherwise same as psTerminated
                result = true;
                break;
             case IntExecutionContext::psRunning:      // can't happen
             case IntExecutionContext::psRunnable:     // can't happen
             case IntExecutionContext::psError:        // unsuccessful execution
             case IntExecutionContext::psFrozen:       // can't happen
                break;
            }
            killTerminatedProcesses();
        }
        catch (std::exception& /*e*/) {
            // Ignore errors.
        }
    }
    return result;
}

void
WGlobalScriptAction::Executer::finish(int /*obj*/, int success)
{
    messageBox(format(_("Command executed successfully on %d unit%!1{s%}."), success), _("Global Actions"));
}
