/**
  *  \file client/cm-distance.cc
  *  \brief Starchart mode: Distance
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/cm-distance.h"
#include "client/chartloc.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "gfx/gfxutil.h"
#include "gfx/palette.h"
#include "ui/event.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/global.h"
#include "game/ship.h"
#include "client/dialogs/helpdlg.h"
#include "game/shippredict.h"

/** Given distance and speed, compute estimated time. */
static int
ETA(double d, int warp)
{
    int32 r = roundToInt(d);
    int32 way = warp*warp;

    return (r + way-1) / way;
}

/** Constructor.
    \param origin Origin of measurement
    \param ship_id Predict fuel usage for this ship (0=none) */
WDistanceChartMode::WDistanceChartMode(GPoint origin, int ship_id)
    : WSimpleChartMode(ChartMode_PrimaryMode),
      origin(origin),
      ship_id(ship_id)
{ }

/** Destructor. */
WDistanceChartMode::~WDistanceChartMode()
{ }

WDistanceChartMode::Mode
WDistanceChartMode::getMode() const
{
    if (ship_id == 0) {
        return DistanceMode;
    } else {
        GShip& sh = getDisplayedTurn().getCurrentUniverse().getShip(ship_id);
        if (sh.isVisible() && sh.getShipKind() == GShip::CurrentShip && sh.getPos() == origin.getSimpleCanonicalLocation()) {
            /* We have a ship and origin is its position */
            if (sh.isPlayable(GObject::Playable)) {
                return WaypointMode;
            } else {
                return ForeignMode;
            }
        } else {
            /* No ship */
            return OtherMode;
        }
    }
}

bool
WDistanceChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Key) {
        const Mode m = getMode();
        switch (event.key.code) {
         case 'x':
            // Swap ends
            {
                GPoint p = origin;
                origin = loc.getLocation();
                loc.setLocation(p);
            }
            return true;

         case 'f':
            // Use ship for fuel consumption
            if (loc.getCurrentShipId() > 0) {
                ship_id = loc.getCurrentShipId();
                loc.markChanged();
            }
            return true;

         case 'w':
         case ss_Ctrl + 'w':
            // Set ship waypoint
            if (m == WaypointMode) {
                GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
                univ.getShip(ship_id).setWaypoint(loc.getLocation().getSimpleNearestAlias(univ.getShip(ship_id).getPos()));
                univ.doScreenUpdates();
            }
            return true;

         case SDLK_BACKSPACE:
            // Back to beginning and exit
            if (m == WaypointMode || m == ForeignMode)
                loc.setCurrentObject(getDisplayedTurn().getCurrentUniverse().getShip(ship_id));
            else
                loc.setLocation(origin);
            delete this;
            return true;

         case SDLK_ESCAPE:
            if (m == WaypointMode || m == ForeignMode)
                loc.setCurrentObject(getDisplayedTurn().getCurrentUniverse().getShip(ship_id));
            delete this;
            return true;

         case 'd':
         case SDLK_DELETE:
         case vk_Quit:
            // Exit mode
            delete this;
            return true;

         case 'h':
         case ss_Alt + 'h':
            doHelp("pcc2:distance");
            return true;
        }
    }
    return WSimpleChartMode::handleEvent(event, loc);
}

void
WDistanceChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    GfxRect pos = loc.getViewport().view;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(COLOR_WHITE);

    const int line = font_heights[FONT_NORMAL];
    const int indent = font_ems[FONT_NORMAL];

    // Distance
    double dist = std::sqrt(double(distanceSquaredRaw(origin, loc.getLocation())));
    outText(ctx, pos.x, pos.y,
            format(_("Distance from first point: %.1f ly"), dist));

    // Speeds, angle
    string_t speed_angle = format(_("Warp/Time: 6/%d 7/%d 8/%d 9/%d"),
                                  ETA(dist, 6),
                                  ETA(dist, 7),
                                  ETA(dist, 8),
                                  ETA(dist, 9));
    if (dist > 0)
        speed_angle += format(", %d" "\xC2\xB0",
                              roundToInt(getHeadingDeg(loc.getLocation().x - origin.x,
                                                       loc.getLocation().y - origin.y)));

    outText(ctx, pos.x, pos.y + line, speed_angle);

    // Ship info
    const Mode m = getMode();
    if (m != DistanceMode) {
        GShip& sh = getDisplayedTurn().getCurrentUniverse().getShip(ship_id);
        if (sh.isPlayed()) {
            /* FIXME: GShipTurnPredictor currently cannot completely deal with ships we do not play. */
            outText(ctx, pos.x, pos.y + 2*line, sh.getName(GObject::LongName) + ":");

            GShipTurnPredictor stp(getDisplayedTurn().getCurrentUniverse(), ship_id);
            stp.setPosition(origin);
            stp.setWaypoint(loc.getLocation());
            stp.computeMovement();

            /* Time and speed */
            string_t text;
            int color;
            if (sh.getWarp() == 0) {
                text = _("not moving");
                color = COLOR_RED;
            } else if (stp.isAtTurnLimit()) {
                text = _("too long");
                color = COLOR_YELLOW;
            } else {
                text = format(_("%d turn%!1{s%} at warp %d"), stp.getNumTurns(), sh.getWarp());
                color = COLOR_WHITE;
            }
            ctx.setColor(color);
            outText(ctx, pos.x + indent, pos.y + 3*line, text);

            /* Fuel */
            if (sh.getWarp() == 0) {
                text = format(_("%d kt fuel aboard"), sh.getCargo(el_Neutronium));
                color = COLOR_WHITE;
            } else {
                text = format(_("%d of %d kt fuel used"), stp.getMovementFuelUsed(), sh.getCargo(el_Neutronium));
                color = stp.getMovementFuelUsed() > sh.getCargo(el_Neutronium) ? COLOR_RED : COLOR_WHITE;
            }
            ctx.setColor(color);
            outText(ctx, pos.x + indent, pos.y + 4*line, text);

            /* Info */
            if (m == WaypointMode) {
                ctx.setColor(COLOR_WHITE);
                outText(ctx, pos.x + indent, pos.y + 5*line, _("Press [W] to change this ship's waypoint."));
            }
        }
    }

    WSimpleChartMode::drawOverlays(can, loc);
}

void
WDistanceChartMode::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_WHITE);

    GfxPoint a = loc.getViewport().scaleXY(origin);
    GfxPoint b = loc.getViewport().scaleXY(loc.getLocation());

    drawLine(ctx, a, b);

    WSimpleChartMode::drawCursor(can, loc);
}

void
WDistanceChartMode::onMove(GChartLocation& loc)
{
    origin += loc.getWrapAdjust();

    WSimpleChartMode::onMove(loc);
}
