/**
  *  \file int/hash.cc
  *  \brief Interpreter: Hashes
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/hash.h"
#include "int/args.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/context.h"
#include "int/vmio.h"
#include "int/propenum.h"

namespace {
    /** Context for iterating a hash. */
    class IntHashContext : public IntContext {
        Ptr<IntHashData> data;
        IntVariableNames::index_t slot;
     public:
        IntHashContext(Ptr<IntHashData> data);
        ~IntHashContext();

        // IntContext:
        int32_t   lookup(const string_t& name);
        void      set(int32_t index, IntValue* value);
        IntValue* get(int32_t index);

        bool      next();
        IntContext* clone();
        GObject*  getObject();
        string_t  toString(bool readable);
        void      store(IntTagNode& sv, Stream& aux);
        void      enumProperties(IntPropertyAcceptor& acceptor);
    };
}


/****************************** IntHashData ******************************/

/** Construct blank hash data. */
IntHashData::IntHashData()
{ }

IntHashData::~IntHashData()
{ }

/** Set element by name. */
void
IntHashData::setNew(const string_t& name, IntValue* value)
{
    IntVariableNames::index_t i;
    try {
        i = names.addMaybe(name);
    }
    catch (...) {
        delete value;
        throw;
    }
    content.setNew(i, value);
}

/** Get element by name. */
IntValue*
IntHashData::get(const string_t& name)
{
    IntVariableNames::index_t i = names.lookup(name);
    if (i != IntVariableNames::nil) {
        return content[i];
    } else {
        return 0;
    }
}

/******************************** IntHash ********************************/

/** Constructor.
    \param data Contained hash data */
IntHash::IntHash(Ptr<IntHashData> data)
    : data(data)
{ }

IntHash::~IntHash()
{ }

IntValue*
IntHash::get(IntArgBlock& args)
{
    string_t key;
    args.checkArgumentCount(1);
    if (!checkStringArg(key, args.getNext())) {
        return 0;
    }
    return IntValue::cloneOf(data->get(key));
}

void
IntHash::set(IntArgBlock& args, IntValue* value)
{
    string_t key;
    args.checkArgumentCount(1);
    if (!checkStringArg(key, args.getNext())) {
        throw IntError::notAssignable();
    }
    data->setNew(key, value->clone());
}

IntContext*
IntHash::makeFirstContext()
{
    if (data->getSize() == 0) {
        return 0;
    } else {
        return new IntHashContext(data);
    }
}

string_t
IntHash::toString(bool /*readable*/)
{
    return "#<hash>";
}

void
IntHash::store(IntTagNode& tag, Stream& /*aux*/)
{
    IntVMSaveContext* vsc = IntVMSaveContext::getCurrentInstance();
    if (vsc != 0) {
        tag.tag   = IntTagNode::Tag_Hash;
        tag.value = vsc->addHash(*data);
    } else {
        throw IntError::notSerializable();
    }
}

IntValue*
IntHash::clone()
{
    return new IntHash(data);
}

/***************************** IntHashContext ****************************/

/** Constructor.
    \param data Hash data. Must contain at least one element. */
IntHashContext::IntHashContext(Ptr<IntHashData> data)
    : data(data),
      slot(0)
{ }

IntHashContext::~IntHashContext()
{ }

int32_t
IntHashContext::lookup(const string_t& name)
{
    if (name == "KEY") {
        /* @q Key:Str (Hash Element Property)
           The key of this hash element. */
        return 0;
    } else if (name == "VALUE") {
        /* @q Value:Any (Hash Element Property)
           The value of this hash element.
           @assignable */
        return 1;
    } else {
        return -1;
    }
}

void
IntHashContext::set(int32_t index, IntValue* value)
{
    if (index == 1) {
        data->setNew(slot, IntValue::cloneOf(value));
    } else {
        throw IntError::notAssignable();
    }
}

IntValue*
IntHashContext::get(int32_t index)
{
    if (index == 0) {
        return makeStringValue(data->getName(slot));
    } else {
        return IntValue::cloneOf(data->get(slot));
    }
}

bool
IntHashContext::next()
{
    if (slot+1 < data->getSize()) {
        ++slot;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntHashContext::clone()
{
    return new IntHashContext(*this);
}

GObject*
IntHashContext::getObject()
{
    return 0;
}

string_t
IntHashContext::toString(bool /*readable*/)
{
    return "#<hashIterator>";
}

void
IntHashContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

void
IntHashContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.addProperty("KEY", thString);
    acceptor.addProperty("VALUE", thNone);
}
