/**
  *  \file int/hash.h
  *  \brief Interpreter: Hashes
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_HASH_H
#define PCC_V2_INT_HASH_H

#include "int/names.h"
#include "int/data.h"
#include "cpluslib/smartptr.h"
#include "int/compound.h"

class IntVMSaveContext;
class IntVMLoadContext;

/** Storage for hash data. Stores keys and values. It relies on
    IntVariableNames to do the actual hashing.

    The exposed interface is a cross of IntVariableNames and
    IntDataSegment. Iteration can therefore simply iterate over
    IntDataSegment indexes. */
class IntHashData {
    friend class IntVMSaveContext;
    friend class IntVMLoadContext;

    IntVariableNames names;
    IntDataSegment content;

 public:
    IntHashData();
    ~IntHashData();

    // Name interface
    void      setNew(const string_t& name, IntValue* value);
    IntValue* get(const string_t& name);

    // Index interface
    void      setNew(IntVariableNames::index_t index, IntValue* value);
    IntValue* get(IntVariableNames::index_t index);
    string_t  getName(IntVariableNames::index_t index);
    IntVariableNames::index_t getSize();
};

/** Hash reference. Hashes are always by-reference, because IntHash
    objects are cloned when put on the stack. The actual data is
    stored in an IntHashData object. */
class IntHash : public IntIndexableValue {
    Ptr<IntHashData> data;

 public:
    IntHash(Ptr<IntHashData> data);
    ~IntHash();

    Ptr<IntHashData> getData();

    // IntIndexableValue:
    IntValue* get(IntArgBlock& args);
    void set(IntArgBlock& args, IntValue* value);
    IntContext* makeFirstContext();

    // IntValue:
    string_t    toString(bool readable);
    void        store(IntTagNode&, Stream&);
    IntValue*   clone();
};


/** Set element by index.
    \param index Index
    \param value Value. Will get owned by IntHashData. */
inline void
IntHashData::setNew(IntVariableNames::index_t index, IntValue* value)
{
    return content.setNew(index, value);
}

/** Get element by index.
    \param index Index
    \return Element, owned by IntHashData */
inline IntValue*
IntHashData::get(IntVariableNames::index_t index)
{
    return content[index];
}

/** Get name by index.
    \param index Index */
inline string_t
IntHashData::getName(IntVariableNames::index_t index)
{
    return names.lookupAddress(index);
}

/** Get number of elements in hash. */
inline IntVariableNames::index_t
IntHashData::getSize()
{
    return names.getNumNames();
}

inline Ptr<IntHashData>
IntHash::getData()
{
    return data;
}

#endif
