/**
  *  \file int/if/richif.cc
  */

#include <climits>
#include <vector>
#include "int/if/richif.h"
#include "cpluslib/string.h"
#include "cpluslib/misc.h"
#include "int/interror.h"
#include "int/args.h"
#include "int/if/ifutil.h"
#include "util/xmlread.h"
#include "ui/richparser.h"
#include "util/misc.h"
#include "io/memstream.h"

/******************************** Internal *******************************/

namespace {
    enum StyleKind {
        kNone,
        kStyle,
        kColor
    };
    struct Style {
        const char* name;
        StyleKind   kind : 8;
        uint8_t     value;
    };
    const Style styles[] = {
        { "",           kNone,  0 },
        { "b",          kStyle, RichTextStyleAttribute::Bold }, /* tag name */
        //{ "background", kColor, UIColor::tc_Background },
        { "big",        kStyle, RichTextStyleAttribute::Big },
        { "blue",       kColor, UIColor::tc_Blue },
        { "bold",       kStyle, RichTextStyleAttribute::Bold }, /* real name */
        //{ "contrast",   kColor, UIColor::tc_Contrast },
        { "dim",        kColor, UIColor::tc_Faded },
        { "em",         kStyle, RichTextStyleAttribute::Bold }, /* tag name (should actually be italic) */
        { "fixed",      kStyle, RichTextStyleAttribute::Fixed }, /* real name */
        { "green",      kColor, UIColor::tc_Green },
        //{ "heading",    kColor, UIColor::tc_Heading },
        //{ "input",      kColor, UIColor::tc_Input },
        //{ "invstatic",  kColor, UIColor::tc_InvStatic },
        //{ "italic",     kStyle, RichTextStyleAttribute::Italic }, /* not supported yet */
        { "kbd",        kStyle, RichTextStyleAttribute::Key }, /* tag name */
        { "key",        kStyle, RichTextStyleAttribute::Key }, /* real name */
        //{ "link",       kColor, UIColor::tc_Link },
        //{ "linkfocus",  kColor, UIColor::tc_LinkFocus },
        //{ "linkshade",  kColor, UIColor::tc_LinkShade },
        { "none",       kNone,  0 },
        { "red",        kColor, UIColor::tc_Red },
        //{ "selection",  kColor, UIColor::tc_Selection },
        { "small"    ,  kStyle, RichTextStyleAttribute::Small },
        { "static",     kColor, UIColor::tc_Static },
        { "tt",         kStyle, RichTextStyleAttribute::Fixed }, /* tag name */
        { "u",          kStyle, RichTextStyleAttribute::Underline }, /* tag name */
        { "underline",  kStyle, RichTextStyleAttribute::Underline }, /* real name */
        { "white",      kColor, UIColor::tc_White },
        { "yellow",     kColor, UIColor::tc_Yellow },
    };
}

static Ptr<RichText>
processStyle(string_t style, Ptr<RichText> text)
{
    style = strTrim(strLCase(style));

    const Style* p = 0;
    for (std::size_t i = 0; i < countof(styles); ++i) {
        if (style == styles[i].name) {
            p = &styles[i];
            break;
        }
    }

    if (!p) {
        throw IntError("Invalid style");
    }

    switch (p->kind) {
     case kNone:
        return text;
     case kColor:
        {
            Ptr<RichText> tmp = new RichText(*text);
            tmp->withAttribute(new RichTextColorAttribute(UIColor::Color(p->value)));
            return tmp;
        }
     case kStyle:
        {
            Ptr<RichText> tmp = new RichText(*text);
            tmp->withAttribute(new RichTextStyleAttribute(RichTextStyleAttribute::Style(p->value)));
            return tmp;
        }
    }

    // Fallback, does not happen
    return text;
}

/**************************** IntRichTextValue ***************************/

IntRichTextValue::IntRichTextValue(Ptr<RichText> value)
    : IntValue(),
      value(value)
{ }

IntRichTextValue::~IntRichTextValue()
{ }

string_t
IntRichTextValue::toString(bool /*readable*/)
{
    return value->getText();
}

void
IntRichTextValue::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

IntValue*
IntRichTextValue::clone()
{
    return new IntRichTextValue(value);
}

Ptr<RichText>
IntRichTextValue::get()
{
    return value;
}

/******************************* Utilities *******************************/

bool
checkRichArg(Ptr<RichText>& out, IntValue* value)
{
    if (value == 0) {
        return false;
    } else {
        if (IntRichTextValue* rv = dynamic_cast<IntRichTextValue*>(value)) {
            out = rv->get();
        } else {
            out = new RichText(value->toString(false));
        }
        return true;
    }
}

/***************************** User Functions ****************************/

/* @q RAdd(args:RichText...):RichText (Function)
   Concatenate all arguments, which can be strings or rich text, to a new rich text string,
   and returns that.

   If any argument is EMPTY, returns EMPTY.
   If no arguments are given, returns an empty (=zero length) rich text string.

   In text mode, this function produces plain strings instead,
   as rich text attributes have no meaning to the text mode applications.

   @since PCC2 1.99.21 */
IntValue*
IFRAdd(IntArgBlock& args)
{
    if (args.getNumArgs() == 1) {
        // Special case: act as cast-to-rich-text, avoiding a copy
        Ptr<RichText> result;
        if (checkRichArg(result, args.getNext())) {
            return new IntRichTextValue(result);
        } else {
            return 0;
        }
    } else {
        // General case
        Ptr<RichText> result = new RichText();
        while (args.getNumArgs() > 0) {
            Ptr<RichText> tmp;
            if (!checkRichArg(tmp, args.getNext())) {
                return 0;
            }
            *result += *tmp;
        }
        return new IntRichTextValue(result);
    }
}

/* @q RMid(str:RichText, first:Int, Optional length:Int):RichText (Function)
   Returns a substring of a rich text string.

   %first specifies the first character position to extract, where 1 means the first.
   %length specifies the number of characters to extract.
   If %length is omitted or EMPTY, the remaining string is extracted.

   If %str or %first are EMPTY, returns EMPTY.

   In text mode, this function deals with plain strings instead,
   as rich text attributes have no meaning to the text mode applications.

   @since PCC2 1.99.21 */
IntValue*
IFRMid(IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    // Parse args
    Ptr<RichText> str;
    int32_t iStart, iLength;
    if (!checkRichArg(str, args.getNext()) || !checkIntArg(iStart, args.getNext(), 0, INT_MAX)) {
        return 0;
    }
    if (!checkIntArg(iLength, args.getNext(), 0, INT_MAX)) {
        iLength = INT_MAX;
    }

    // Convert BASIC indexes to C++ indexes
    string_t::size_type nStart  = iStart == 0 ? 0 : static_cast<string_t::size_type>(iStart) - 1;
    string_t::size_type nLength =                   static_cast<string_t::size_type>(iLength);
    if (nStart > str->size()) {
        return new IntRichTextValue(new RichText());
    } else {
        return new IntRichTextValue(new RichText(str->substr(nStart, nLength)));
    }
}

/* @q RString(str:RichText):Str (Function)
   Returns the text content of a rich text string,
   i.e. the string with all attributes removed.

   If %str is EMPTY, returns EMPTY.

   In text mode, this function deals with plain strings instead,
   as rich text attributes have no meaning to the text mode applications.

   @since PCC2 1.99.21 */
IntValue*
IFRString(IntArgBlock& args)
{
    args.checkArgumentCount(1);
    Ptr<RichText> str;
    if (checkRichArg(str, args.getNext())) {
        return makeStringValue(str->getText());
    } else {
        return 0;
    }
}

/* @q RLen(str:RichText):Int (Function)
   Returns the number of characters in a rich text string.

   If %str is EMPTY, returns EMPTY.

   In text mode, this function deals with plain strings instead,
   as rich text attributes have no meaning to the text mode applications.

   @since PCC2 1.99.21 */
IntValue*
IFRLen(IntArgBlock& args)
{
    args.checkArgumentCount(1);
    Ptr<RichText> str;
    if (checkRichArg(str, args.getNext())) {
        return makeIntValue(str->size());
    } else {
        return 0;
    }
}

/* @q RStyle(style:Str, content:RichText...):RichText (Function)
   Attaches a new style to a rich text string.
   Concatenates all %content parameters, and returns a new rich text string with the specified attribute added.

   <pre class="ccscript">
     RStyle("red", "This is ", RStyle("bold", "great"))
   </pre>
   produces "<font color="red">This is <b>great</b></font>".

   If any argument is EMPTY, returns EMPTY.

   In text mode, this function just returns the concatenation of the %content,
   as rich text attributes have no meaning to the text mode applications.

   @todo document the styles
   @since PCC2 1.99.21
   @see RLink, RXml */
IntValue*
IFRStyle(IntArgBlock& args)
{
    // Read style argument
    args.checkArgumentCountAtLeast(1);
    string_t style;
    if (!checkStringArg(style, args.getNext())) {
        return 0;
    }

    // Read remaining arguments, converting them to Rich Text. This is just what IFRAdd does.
    Ptr<RichText> result;
    if (!checkRichArg(result, IFRAdd(args))) {
        return 0;
    }

    // Process the style
    string_t::size_type pos = 0, i;
    while ((i = style.find(',', pos)) != string_t::npos) {
        result = processStyle(style.substr(pos, i-pos), result);
        pos = i+1;
    }
    return new IntRichTextValue(processStyle(style.substr(pos), result));
}

/* @q RLink(target:Str, content:RichText...):RichText (Function)
   Attaches a link to a rich text string.
   Produces a rich text string that contains a link to the specified target,
   and the concatenation of all %content parameters as text.

   If any argument is EMPTY, returns EMPTY.

   In text mode, this function just returns the concatenation of the %content,
   as rich text attributes have no meaning to the text mode applications.

   @since PCC2 1.99.21
   @see RStyle, RXml */
IntValue*
IFRLink(IntArgBlock& args)
{
    // Read link argument
    args.checkArgumentCountAtLeast(1);
    string_t link;
    if (!checkStringArg(link, args.getNext())) {
        return 0;
    }

    // Read remaining arguments, converting them to Rich Text. This is just what IFRAdd does.
    Ptr<RichText> result;
    if (!checkRichArg(result, IFRAdd(args))) {
        return 0;
    }

    // Build a link
    Ptr<RichText> clone = new RichText(*result);
    clone->withAttribute(new RichTextLinkAttribute(link));
    return new IntRichTextValue(clone);
}

namespace {
    class IntXmlReader : public XmlReader {
     public:
        IntXmlReader(Stream& s)
            : XmlReader(s),
              args()
            { }

        void addArg(IntValue* value)
            { args.push_back(value); }

        virtual string_t expandEntity(string_t e);

     private:
        std::vector<IntValue*> args;
    };

    string_t
    IntXmlReader::expandEntity(string_t e)
    {
        int n;
        if (stringValue(e, n)) {
            // Completely numeric
            if (n >= 0 && n < int(args.size())) {
                return IntValue::toStringOf(args[n], false);
            } else {
                return string_t();
            }
        } else {
            // Default handling
            return XmlReader::expandEntity(e);
        }
    }
}

/* @q RXml(xml:Str, args:Str...):RichText (Function)
   Create rich text string from XML.
   Parses the %xml string.
   Tags are converted into rich text attributes.
   Entity references of the form &amp;&lt;digits&gt;; are replaced by the respective element from %args,
   where the first element is &amp;0;.

   For example,
   <pre class="ccscript">
     RXml("&lt;font color='&0;'>This is &lt;b>&1;&lt;/b>&lt;/font>", "red", "great")
   </pre>
   produces <font color="red">This is <b>great</b></font>.

   In text mode, this function uses a simpler XML parser, and returns a plain string,
   as rich text attributes have no meaning to the text mode applications.

   @todo document the styles
   @since PCC2 1.99.21
   @see RStyle, RLink */
IntValue*
IFRXml(IntArgBlock& args)
{
    // Read XML text
    args.checkArgumentCountAtLeast(1);
    string_t xml;
    if (!checkStringArg(xml, args.getNext())) {
        return 0;
    }

    // Construct XML reader
    MemoryStream ms(xml);
    IntXmlReader rdr(ms);
    while (args.getNumArgs() > 0) {
        rdr.addArg(args.getNext());
    }
    RichParser p(rdr);
    p.next();

    // Read
    Ptr<RichText> result = new RichText(p.parse());
    return new IntRichTextValue(result);
}
