/**
  *  \file int/if/specif.cc
  *  \brief Interpreter Game Interface: Specification Access
  *
  *  This implements access to Torpedo, Beam, and Engine specifications.
  *
  *  \todo It's pretty ugly that we have three almost identical classes here.
  *  Can we do better?
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/specif.h"
#include "int/value.h"
#include "game/spec.h"
#include "game/host.h"
#include "game/config.h"
#include "int/if/ifutil.h"
#include "cpluslib/misc.h"
#include "int/interror.h"
#include "int/data.h"
#include "int/compound.h"
#include "int/args.h"
#include "cpluslib/format.h"
#include "int/contextio.h"
#include "int/propenum.h"

/**********************                             **********************/
/********************** Part I: Interface Functions **********************/
/**********************                             **********************/

/* @q Torpedo(id:Int):Obj (Function, Context)
   Access torpedo properties.
   Use as
   | ForEach Torpedo Do ...
   or
   | With Torpedo(n) Do ...

   @diff This function was available for use in %With under the name %Torpedoes() since PCC 1.0.6.
   Do not use the name %Torpedoes in new code, it is not supported by PCC2; use %Torpedo instead.

   @see int:index:group:torpedoproperty|Torpedo Properties, Launcher()
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFTorpedoGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_TORPS))
        return 0;
    return new IntTorpedoContext(id);
}

IntContext*
IFTorpedoMake()
{
    return new IntTorpedoContext(1);
}

/* @q Launcher(id:Int):Obj (Function, Context)
   Access torpedo launcher properties.
   Use as
   | ForEach Launcher Do ...
   or
   | With Launcher(n) Do ...

   @diff This function was available for use in %With under the name %Launchers() since PCC 1.0.6.
   Do not use the name %Launchers in new code, it is not supported by PCC2; use %Launcher instead.

   @see int:index:group:torpedoproperty|Torpedo Properties, Torpedo()
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFLauncherGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_TORPS))
        return 0;
    return new IntLauncherContext(id);
}

IntContext*
IFLauncherMake()
{
    return new IntLauncherContext(1);
}

/* @q Beam(id:Int):Obj (Function, Context)
   Access beam weapon properties.
   Use as
   | ForEach Beam Do ...
   or
   | With Beam(n) Do ...

   @diff This function was available for use in %With under the name %Beams() since PCC 1.0.6.
   Do not use the name %Beams in new code, it is not supported by PCC2; use %Beam instead.

   @see int:index:group:beamproperty|Beam Properties
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFBeamGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_BEAMS))
        return 0;
    return new IntBeamContext(id);
}

IntContext*
IFBeamMake()
{
    return new IntBeamContext(1);
}

/* @q Engine(id:Int):Obj (Function, Context)
   Access engine weapon properties.
   Use as
   | ForEach Engine Do ...
   or
   | With Engine(n) Do ...

   @diff This function was available for use in %With under the name %Engines() since PCC 1.0.6.
   Do not use the name %Engines in new code, it is not supported by PCC2; use %Engine instead.

   @see int:index:group:engineproperty|Engine Properties
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFEngineGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_ENGINES))
        return 0;
    return new IntEngineContext(id);
}

IntContext*
IFEngineMake()
{
    return new IntEngineContext(1);
}

/***************************                   ***************************/
/*************************** Part II: Contexts ***************************/
/***************************                   ***************************/

namespace {
    enum { ComponentProperty, WeaponProperty, EngineProperty };
}

// Mapping for torpedoes and launchers:
static const TIntMapping torpedo_map[] = {
    { "COST.D",        icpCostD,     ComponentProperty, thInt },
    { "COST.M",        icpCostM,     ComponentProperty, thInt },
    { "COST.MC",       icpCostMC,    ComponentProperty, thInt },
    { "COST.STR",      icpCostStr,   ComponentProperty, thString },
    { "COST.T",        icpCostT,     ComponentProperty, thInt },
    { "DAMAGE",        iwpDamage,    WeaponProperty,    thInt },
    { "ID",            icpId,        ComponentProperty, thInt },
    { "KILL",          iwpKill,      WeaponProperty,    thInt },
    { "MASS",          icpMass,      ComponentProperty, thInt },
    { "NAME",          icpName,      ComponentProperty, thString },
    { "NAME.SHORT",    icpNameShort, ComponentProperty, thString },
    { "TECH",          icpTech,      ComponentProperty, thInt },
    { "TECH.TORPEDO",  icpTech,      ComponentProperty, thInt },
};

// Mapping for beams:
static const TIntMapping beam_map[] = {
    { "COST.D",        icpCostD,     ComponentProperty, thInt },
    { "COST.M",        icpCostM,     ComponentProperty, thInt },
    { "COST.MC",       icpCostMC,    ComponentProperty, thInt },
    { "COST.STR",      icpCostStr,   ComponentProperty, thString },
    { "COST.T",        icpCostT,     ComponentProperty, thInt },
    { "DAMAGE",        iwpDamage,    WeaponProperty,    thInt },
    { "ID",            icpId,        ComponentProperty, thInt },
    { "KILL",          iwpKill,      WeaponProperty,    thInt },
    { "MASS",          icpMass,      ComponentProperty, thInt },
    { "NAME",          icpName,      ComponentProperty, thString },
    { "NAME.SHORT",    icpNameShort, ComponentProperty, thString },
    { "TECH",          icpTech,      ComponentProperty, thInt },
    { "TECH.BEAM",     icpTech,      ComponentProperty, thInt },
};

// Mapping for engines:
static const TIntMapping engine_map[] = {
    { "COST.D",        icpCostD,         ComponentProperty, thInt },
    { "COST.M",        icpCostM,         ComponentProperty, thInt },
    { "COST.MC",       icpCostMC,        ComponentProperty, thInt },
    { "COST.STR",      icpCostStr,       ComponentProperty, thString },
    { "COST.T",        icpCostT,         ComponentProperty, thInt },
    { "FUELFACTOR",    iepFuelFactor,    EngineProperty,    thArray },
    { "ID",            icpId,            ComponentProperty, thInt },
    { "NAME",          icpName,          ComponentProperty, thString },
    { "NAME.SHORT",    icpNameShort,     ComponentProperty, thString },
    { "SPEED$",        iepEfficientWarp, EngineProperty,    thInt },
    { "TECH",          icpTech,          ComponentProperty, thInt },
    { "TECH.ENGINE",   icpTech,          ComponentProperty, thInt },
};

/*************************** IntTorpedoContext ***************************/

IntTorpedoContext::IntTorpedoContext(int id)
    : id(id)
{ }

IntTorpedoContext::~IntTorpedoContext()
{ }

int32_t
IntTorpedoContext::lookup(const string_t& name)
{
    return lookupName(name, torpedo_map, countof(torpedo_map));
}

void
IntTorpedoContext::set(int32_t index, IntValue* value)
{
    switch (torpedo_map[index].domain) {
     case ComponentProperty:
        // Although we're dealing with the Torpedo here, this uses the launcher.
        // The launcher is the long-lived object, the torpedo is the short-lived one
        // which makes no sense to modify.
        setComponentProperty(getTorp(id), IntComponentProperty(torpedo_map[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntTorpedoContext::get(int32_t index)
{
    switch (torpedo_map[index].domain) {
     case ComponentProperty: return getComponentProperty(getTorp(id).getTorpedo(), IntComponentProperty(torpedo_map[index].index));
     case WeaponProperty:    return getWeaponProperty   (getTorp(id).getTorpedo(), IntWeaponProperty   (torpedo_map[index].index), true);
    }
    return 0;
}

bool
IntTorpedoContext::next()
{
    if (id < NUM_TORPS) {
        ++id;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntTorpedoContext::clone()
{
    return new IntTorpedoContext(id);
}

GObject*
IntTorpedoContext::getObject()
{
    return 0;
}

string_t
IntTorpedoContext::toString(bool /*readable*/)
{
    return format("Torpedo(%d)", id);
}

void
IntTorpedoContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Torpedo;
    sv.value = id;
}

void
IntTorpedoContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(torpedo_map, countof(torpedo_map));
}

/*************************** IntLauncherContext **************************/

IntLauncherContext::IntLauncherContext(int id)
    : id(id)
{ }

IntLauncherContext::~IntLauncherContext()
{ }

int32_t
IntLauncherContext::lookup(const string_t& name)
{
    return lookupName(name, torpedo_map, countof(torpedo_map));
}

void
IntLauncherContext::set(int32_t index, IntValue* value)
{
    switch (torpedo_map[index].domain) {
     case ComponentProperty:
        setComponentProperty(getTorp(id), IntComponentProperty(torpedo_map[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntLauncherContext::get(int32_t index)
{
    switch (torpedo_map[index].domain) {
     case ComponentProperty: return getComponentProperty(getTorp(id), IntComponentProperty(torpedo_map[index].index));
     case WeaponProperty:    return getWeaponProperty   (getTorp(id), IntWeaponProperty   (torpedo_map[index].index), true);
    }
    return 0;
}

bool
IntLauncherContext::next()
{
    if (id < NUM_TORPS) {
        ++id;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntLauncherContext::clone()
{
    return new IntLauncherContext(id);
}

GObject*
IntLauncherContext::getObject()
{
    return 0;
}

string_t
IntLauncherContext::toString(bool /*readable*/)
{
    return format("Launcher(%d)", id);
}

void
IntLauncherContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Launcher;
    sv.value = id;
}

void
IntLauncherContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(torpedo_map, countof(torpedo_map));
}

/***************************** IntBeamContext ****************************/

IntBeamContext::IntBeamContext(int id)
    : id(id)
{ }

IntBeamContext::~IntBeamContext()
{ }

int32_t
IntBeamContext::lookup(const string_t& name)
{
    return lookupName(name, beam_map, countof(beam_map));
}

void
IntBeamContext::set(int32_t index, IntValue* value)
{
    switch (beam_map[index].domain) {
     case ComponentProperty:
        setComponentProperty(getBeam(id), IntComponentProperty(beam_map[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntBeamContext::get(int32_t index)
{
    switch (beam_map[index].domain) {
     case ComponentProperty: return getComponentProperty(getBeam(id), IntComponentProperty(beam_map[index].index));
     case WeaponProperty:    return getWeaponProperty   (getBeam(id), IntWeaponProperty   (beam_map[index].index), false);
    }
    return 0;
}

bool
IntBeamContext::next()
{
    if (id < NUM_BEAMS) {
        ++id;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntBeamContext::clone()
{
    return new IntBeamContext(id);
}

GObject*
IntBeamContext::getObject()
{
    return 0;
}

string_t
IntBeamContext::toString(bool /*readable*/)
{
    return format("Beam(%d)", id);
}

void
IntBeamContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Beam;
    sv.value = id;
}

void
IntBeamContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(beam_map, countof(beam_map));
}

/************************** IntEngineFuelFactor **************************/

namespace {
    /** Implementation of Engine().FuelFactor(). */
    class IntEngineFuelFactor : public IntIndexableValue {
        const GEngine& e;
     public:
        IntEngineFuelFactor(const GEngine& e)
            : e(e)
            { }
        ~IntEngineFuelFactor()
            { }
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& /*args*/, IntValue* /*value*/)
            { throw IntError::notAssignable(); }
        IntContext* makeFirstContext()
            { throw IntError::typeError(IntError::tyIterable); }
        string_t toString(bool /*readable*/)
            { return "#<array>"; }
        void store(IntTagNode&, Stream&)
            { throw IntError::notSerializable(); }
        IntValue* clone()
            { return new IntEngineFuelFactor(e); }
    };
}

IntValue*
IntEngineFuelFactor::get(IntArgBlock& args)
{
    int32_t warp;
    args.checkArgumentCount(1);
    if (!checkIntArg(warp, args.getNext(), 0, 9))
        return 0;
    return makeIntValue(e.getFuelFactor(warp));
}

/**************************** IntEngineContext ***************************/

IntEngineContext::IntEngineContext(int id)
    : id(id)
{ }

IntEngineContext::~IntEngineContext()
{ }

int32_t
IntEngineContext::lookup(const string_t& name)
{
    return lookupName(name, engine_map, countof(engine_map));
}

void
IntEngineContext::set(int32_t index, IntValue* value)
{
    switch (engine_map[index].domain) {
     case ComponentProperty:
        setComponentProperty(getEngine(id), IntComponentProperty(engine_map[index].index), value);
        break;
     case EngineProperty:
        setEngineProperty(getEngine(id), IntEngineProperty(engine_map[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntEngineContext::get(int32_t index)
{
    switch (engine_map[index].domain) {
     case ComponentProperty: return getComponentProperty(getEngine(id), IntComponentProperty(engine_map[index].index));
     case EngineProperty:    return getEngineProperty   (getEngine(id), IntEngineProperty   (engine_map[index].index));
    }
    return 0;
}

bool
IntEngineContext::next()
{
    if (id < NUM_ENGINES) {
        ++id;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntEngineContext::clone()
{
    return new IntEngineContext(id);
}

GObject*
IntEngineContext::getObject()
{
    return 0;
}

string_t
IntEngineContext::toString(bool /*readable*/)
{
    return format("Engine(%d)", id);
}

void
IntEngineContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Engine;
    sv.value = id;
}

void
IntEngineContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(engine_map, countof(engine_map));
}

/***********************                           ***********************/
/*********************** Part III: Property Access ***********************/
/***********************                           ***********************/

/** Get property of a component.
    \param comp Component to query
    \param icp  Property to query */
IntValue*
getComponentProperty(const GComponent& comp, IntComponentProperty icp)
{
    switch (icp) {
     case icpMass:
        /* @q Mass:Int (Beam Property, Torpedo Property)
           Mass of this component, in kt. */
        return makeIntValue(comp.getMass());
     case icpTech:
        /* @q Tech:Int (Hull Property, Engine Property, Beam Property, Torpedo Property)
           @q Tech.Engine:Int (Engine Property)
           @q Tech.Beam:Int (Beam Property)
           @q Tech.Torpedo:Int (Torpedo Property)
           Tech level of this component. */
        /* @q Tech.Hull:Int (Hull Property, Ship Property)
           Hull tech level. */
        return makeIntValue(comp.getTech());
     case icpCostT:
        /* @q Cost.T:Int (Hull Property, Engine Property, Beam Property, Torpedo Property)
           Tritanium cost of this component. */
        return makeIntValue(comp.getCost().get(el_Tritanium));
     case icpCostD:
        /* @q Cost.D:Int (Hull Property, Engine Property, Beam Property, Torpedo Property)
           Duranium cost of this component. */
        return makeIntValue(comp.getCost().get(el_Duranium));
     case icpCostM:
        /* @q Cost.M:Int (Hull Property, Engine Property, Beam Property, Torpedo Property)
           Molybdenum cost of this component. */
        return makeIntValue(comp.getCost().get(el_Molybdenum));
     case icpCostMC:
        /* @q Cost.MC:Int (Hull Property, Engine Property, Beam Property, Torpedo Property)
           Money cost of this component. */
        return makeIntValue(comp.getCost().get(el_Money));
     case icpCostSup:
        // FIXME: not referenced!
        return makeIntValue(comp.getCost().get(el_Supplies));
     case icpCostStr:
        /* @q Cost.Str:Cargo (Hull Property, Engine Property, Beam Property, Torpedo Property)
           Cost of this component, as a string. */
        return makeStringValue(comp.getCost().toCargospec());
     case icpName:
        /* @q Name:Str (Hull Property, Engine Property, Beam Property, Torpedo Property)
           Name of this component.
           @assignable */
        /* @q Hull:Str (Ship Property)
           Name of the ship's hull. */
        return makeStringValue(comp.getName());
     case icpNameShort:
        /* @q Name.Short:Str (Hull Property, Engine Property, Beam Property, Torpedo Property)
           Short name of this component.
           @assignable */
        /* @q Hull.Short:Str (Ship Property)
           Short name of the hull. */
        return makeStringValue(comp.getShortName());
     case icpId:
        /* @q Id:Str (Hull Property, Engine Property, Beam Property, Torpedo Property)
           Component Id. */
        /* @q Hull$:Str (Ship Property)
           Hull Id. */
        return makeIntValue(comp.getId());
    }
    return 0;
}

void
setComponentProperty(GComponent& comp, IntComponentProperty icp, IntValue* value)
{
    string_t n;
    switch (icp) {
     case icpName:
        if (checkStringArg(n, value)) {
            comp.setName(n);
        }
        break;
     case icpNameShort:
        if (checkStringArg(n, value)) {
            comp.setShortName(n);
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

/** Get property of a weapon. */
IntValue*
getWeaponProperty(const GWeapon& w, IntWeaponProperty iwp, bool isTorpedo)
{
    /* We're doubling torpedo bang/kill values when appropriate */
    int factor;
    if (isTorpedo && (host.isTHost() || !config.AllowAlternativeCombat()))
        factor = 2;
    else
        factor = 1;

    switch (iwp) {
     case iwpKill:
        /* @q Kill:Int (Beam Property, Torpedo Property)
           Anti-life power of this weapon. */
        return makeIntValue(factor * w.getKill());
     case iwpDamage:
        /* @q Damage:Int (Beam Property, Torpedo Property)
           Explosive power of this weapon. */
        return makeIntValue(factor * w.getDamage());
    }
    return 0;
}

/** Get property of an engine. */
IntValue*
getEngineProperty(const GEngine& e, IntEngineProperty iep)
{
    switch (iep) {
     case iepEfficientWarp:
        /* @q Speed$:Int (Engine Property)
           Nominal speed of this engine.
           This is the speed PCC considers "optimal" for this engine.
           It defaults to the lowest speed at which the engine runs at 120% fuel consumption or less.
           You can assign a value between 1 and 9 to this property to change what PCC considers optimal.
           @since PCC 1.1.15, PCC2 1.99.8 */
        return makeIntValue(e.getMaxEfficientWarp());
     case iepFuelFactor:
        /* @q FuelFactor:Int() (Engine Property)
           Array of fuel factors for warp factors from 0 to 9.
           This value is used in the computation of fuel usage. */
        return new IntEngineFuelFactor(e);
    }
    return 0;
}

/** Set property of an engine. */
void
setEngineProperty(GEngine& e, IntEngineProperty iep, IntValue* value)
{
    int32_t val;
    switch (iep) {
     case iepEfficientWarp:
        if (checkIntArg(val, value, 0, 9))
            e.setMaxEfficientWarp(val);
        break;
     default:
        throw IntError::notAssignable();
    }
}
