/**
  *  \file ui/combobox.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/combobox.h"
#include "util/misc.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/util.h"
#include "ui/widgets/focusgroup.h"
#include "ui/widgets/button.h"

/** Create combo box.
    \param id      Widget id
    \param min,max Value range
    \param list    List of names for values. Should contain one name for
                   every value in range [min,max]. */
UIComboBox::UIComboBox(int id, int min, int max, const UIStringList& list)
    : UINumberSelector(id, min, max, 1),
      list(list)
{ }

UIComboBox::~UIComboBox()
{ }

/** Draw. */
void
UIComboBox::drawContent(GfxCanvas& can)
{
    UIStringList::iterator i = list.getItemByKey(getValue());
    string_t text;
    if (i == list.end())
        text = numToString(getValue());
    else
        text = i->second;

    // FIXME: skin
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    drawSolidBar(can, extent, standard_colors[COLOR_GRAY]);
    if (getFocusState() != NoFocus)
        ctx.setColor(COLOR_BLUE);
    else
        ctx.setColor(COLOR_BLACK);
    ctx.useFont(*fonts[FONT_TITLE]);
    outTextF(ctx, extent, text);
    can.update(getExtent());
}

/** Show popup menu of all possible values. */
void
UIComboBox::popupMenu()
{
    UIStandardListbox the_list_box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 30, list);
    the_list_box.scrollToKey(getValue());
    if (the_list_box.doStandardMenu(*this))
        setValue(the_list_box.getKey());
}

void
UIComboBox::getLayoutInfo(LayoutInfo& li)
{
    li.min_size = li.pref_size = GfxPoint(getTextWidth(*fonts[FONT_TITLE], list), font_heights[FONT_TITLE]);
    li.flags = li_FlexibleH;
}

bool
UIComboBox::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    // If we want to handle cm_INT_QueryReferencePoint, we can do that here.
    // Right now, the default implementation suffices.
    return UINumberSelector::handleCommand(cmd, sender);
}

/** Create combo box controls into a container.
    \param container container to receive the new widget
    \param box       combo box, already created, will be added into new structure
    \param h         space to store objects */
UIComboBox&
createComboBoxControls(UIGroup& container, UIComboBox& box, ObjectHolder& h)
{
    UIFocusGroup& frame_group = h.add(new UIFocusGroup(UIHBoxLayout::instance5, 0, 5));
    UIButton& button_dec = h.add(new UIButton(0, "-", '-'));
    UIButton& button_inc = h.add(new UIButton(0, "+", '+'));
    UIButton& button_tab = h.add(new UIButton(0, "\xEE\x85\x80", SDLK_TAB));

    frame_group.add(button_dec);
    frame_group.add(box);
    frame_group.add(button_tab);
    frame_group.add(button_inc);

    button_inc.sig_fired.add(static_cast<UINumberSelector*>(&box), &UINumberSelector::increment);
    button_dec.sig_fired.add(static_cast<UINumberSelector*>(&box), &UINumberSelector::decrement);
    button_tab.sig_fired.add(&box, (void (UIComboBox::*)()) &UIComboBox::popupMenu);

    container.add(frame_group);

    return box;
}
