/**
  *  \file ui/uiglobal.cc
  *  \brief User Interface Global Variables
  *
  *  Global Variables and initialisation routines thereof.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdio>
#include <vector>
#include "ui/uiglobal.h"
#include "util/console.h"
#include "gfx/pixmap.h"
#include "ui/event.h"
#include "io/streams.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "resmgr/resmgr.h"
#include "util/except.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "util/cmdline.h"
#include "gfx/init.h"
#include "io/dirs.h"
#include "resmgr/resdir.h"
#include "arch/platform.h"

/** All the fonts. Fonts are loaded at startup. */
Ptr<GfxFont> fonts[MAX_FONTS];

/** Font heights. Cached here for simplicity. */
int font_heights[MAX_FONTS];

/** "em" widths for all fonts. Cached here for simplicity. The "em"
    width is the width of the letter "m" and is used as an estimate
    for font-dependant width computations.

    For "normal-height" fonts (16px and above, i.e. FONT_NORMAL,
    FONT_TITLE), we have the following empirical measures:
    - The average width of a character is about 60% em. Thus, as a
      rule of thumb, three characters usually fit into two em.
    - The maximum width of a character is about 100% em. Thus, three
      characters will, with most fonts, always fit into three em,
      but usually waste screen real-estate (some fonts have larger
      "W" or "M").

    "Small" fonts (e.g. 10px, FONT_SMALL) are generally wider for
    readability. Here, we have:
    - The average width is often 80..100% em.
    - The maximum width is often as large as 150% em. */
int font_ems[MAX_FONTS];

GfxPoint font_sizes[MAX_FONTS];

/** Root widget. */
Ptr<UIRoot> ui_root;

/** Screen surface. This is the surface on which all user interface
    activities happen. */
Ptr<GfxSurface> ui_screen;

/** Bit depth of screen surface. Cached here for simplicity. */
int ui_screen_bits;

/** Resource directory. */
static Ptr<Directory> resource_directory;

/** Additional resources, added using --resource. */
static std::vector<string_t> additional_resources;

static bool
loadFont(int i)
{
    try {
        Ptr<GfxFont> fnt = ResManager::getInstance().getFont(ResId("font", i));
        if (!fnt)
            return false;

        fonts[i-1] = fnt;
        font_heights[i-1] = fonts[i-1]->getTextHeight("Tp");
        font_ems[i-1] = fonts[i-1]->getTextWidth("m");
        font_sizes[i-1] = GfxPoint(font_ems[i-1], font_heights[i-1]);
        return true;
    }
    catch(...) {
        return false;
    }
}

static void
optAddResource(StandardCommandLineParser& p, const StandardCommandLineParser::Option&)
{
    additional_resources.push_back(p.getArg());
}

static void
optSetNoMouseGrab(StandardCommandLineParser&, const StandardCommandLineParser::Option&)
{
    setInhibitMouseGrab(true);
}

void
initUIOptions(StandardCommandLineParser& p)
{
    static const StandardCommandLineParser::Option stuff[] = {
        { "resource", &optAddResource, 0, true,
          N_(" -resource=NAME     Add resource provider") },
        { "nomousegrab", &optSetNoMouseGrab, 0, false,
          N_(" -nomousegrab       Don't grab (lock into window) mouse pointer") }
    };
    p.addOptionSet(stuff, countof(stuff));
}

/** Initialize. This sets up all GUI-related variables.
    \param iconName Name of icon (file)
    \param appName  Name of application, for window manager title */
void
initUIGlobals(const char* iconName, const char* appName)
{
    // Initialize graphics
    SDL_Surface* sfc = initGraphics(iconName, appName);
    if (!sfc)
        throw GraphicsIOException(_("Graphics initialisation failed."));
    Ptr<GfxSurface> screen = new GfxSurface(sfc, false);

    console.write(LOG_INFO, _("=== Initializing GUI... ==="));

    char driver_name[100];
    if (SDL_VideoDriverName(driver_name, sizeof driver_name)) {
        const SDL_VideoInfo* vi = SDL_GetVideoInfo();
        string_t flags;
        if (vi->wm_available)   // do not translate the flag strings
            flags += ", wm";
        if (vi->hw_available)
            flags += ", hw";
        if (vi->blit_hw)
            flags += ", hw->hw";
        if (vi->blit_sw)
            flags += ", sw->hw";
        if (vi->blit_fill)
            flags += ", fill";
        console.write(LOG_INFO, format(_("Video driver: %s (%dk%s)"),
                                       driver_name, vi->video_mem, flags));
    }

    /* initialize UI */
    initPalette(sfc);

    resource_directory = Directory::openDir(Platform::getResourceDirectory());
    ui_screen = screen;
    ui_screen_bits = screen->getBitsPerPixel();
    ui_root = new UIRoot(screen);

    ResManager::getInstance().addNewProvider(new ResDirectory(resource_directory));

    /* initialize resources */
    for (std::vector<string_t>::iterator i = additional_resources.begin(); i != additional_resources.end(); ++i)
        loadResource(ResManager::getInstance(), *i);

    try {
        Ptr<Stream> s = resource_directory->openFile("cc-res.cfg", Stream::C_READ);
        loadResourceConfig(ResManager::getInstance(), *s);
    }
    catch(...) { }

    try {
        Ptr<Stream> s = Stream::openFile(Directory::makeFileName(Platform::getUserSettingsDirectory(1), "cc-res.cfg"), Stream::C_READ);
        loadResourceConfig(ResManager::getInstance(), *s);
    }
    catch (...) { }

    /* load fonts */
    int i = 1;
    while (loadFont(i))
        ++i;
    console.write(LOG_INFO, format(_("%d font%!1{s%} loaded"), i-1));
    if (i <= MIN_FONTS)
        throw GraphicsIOException(format(_("Too few fonts available, need at least %d"), int(MIN_FONTS)));

    initEvents();
}

/** Post a timer event. Call this from long computations to update
    the background.
    \see UIRoot::postTimerEvent() */
void
postTimerEvent()
{
    if (ui_root)
        ui_root->postTimerEvent();
}

/** Take a screenshot. Screenshot will be saved in current directory. */
void
doScreenshot()
{
    // FIXME: this is pretty C-ish.
    using namespace std;
    FILE* fp;
    int index = 0;
    char fname[100];
    while (1) {
        sprintf(fname, "file%04d.bmp", ++index);
        if ((fp = fopen(fname, "rb")) == 0)
            break;
        fclose(fp);
        if (index >= 9999) {
            console.write(LOG_ERROR,
                          _("Unable to find a free file name for screenshot"));
            return;
        }
    }
    SDL_SaveBMP(ui_screen->getSystemSurface(), fname);
    console.write(LOG_INFO, format(_("Screenshot saved as `%s'."), fname));
}
